#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include "dr-file.h"
#include "xournal.h"

#include <string.h>
#include <gtk/gtk.h>
#include <sys/stat.h>
#include <unistd.h> // sysconf
#include <locale.h>
#include <liberutils/er_error.h>
#include <libermetadb/ermetadb.h>

#define TEMP_PATH "/media/mmcblk0p1/Programs"

#define METADATA_KEY_XOURNAL_LAST_PAGE  "xournal_last_read_location"

#define MOUNTPOINT_CARD   "/media/mmcblk0p1"

static void write_document_info_to_database (erMetadb db, const gchar *filename, int last_page);
static int read_document_info_from_database (erMetadb db, const gchar *filename);


// writing and reading a temp file.
gboolean
save_to_temp_file(void)
{
  	gchar *tmpfn;
  	gboolean success;

  	end_text();
  	reset_focus();
	// This function is called when the DR is shutting down, so save the position 
	// of the current document
	// Store the last page
	if (ui.filename != NULL)
	{
		add_last_page_to_meta_data(ui.filename, ui.pageno);
	}


	// Some people report that auto shutdown can corrupt files, so never save to
	// existing file
  	//if (ui.filename == NULL) 
	{
	  	tmpfn = g_build_filename(TEMP_PATH, ".xournal_sav.xoj", NULL);
  	}
	//else
	//{
	//	tmpfn = g_strdup(ui.filename);
	//}
	
  	// overwrite any exsisting!
  	success = save_journal(tmpfn);
  	// we can not take any action, since we are running in background or
  	// we are shutting down
	
  	g_free(tmpfn);
	
  	return success;
}

gboolean
read_from_temp_file(void)
{
    gchar *tmpfn;
    gboolean success;

    tmpfn = g_build_filename(TEMP_PATH, ".xournal_sav.xoj", NULL);

    success = open_journal(tmpfn);
    if (success)
    {
	    // The file exsisted and is now read, so remove it.
	    g_remove(tmpfn);
    }
	
    g_free(tmpfn);
	
    return success;
}

gboolean
write_scribbles_to_file(const char *filename)
{
	FILE *f;
	GList *pagelist, *layerlist, *itemlist, *list;
	struct Page *pg, *tmppg;
	struct Layer *layer;
	struct Item *item;
	int page_num;
	int layer_num;
	int i;
	
	f = fopen(filename, "w");
	if (f==NULL) return FALSE;
	chk_attach_names();

	setlocale(LC_NUMERIC, "C");

	// Create the header for the file
	fprintf(f, "<?xml version=\"1.0\" encoding=\"utf-8\"?>\n");
	fprintf(f, "<notes>\n"); // notes
	fprintf(f, "<version><number>1.0</number><organisation>xournal</organisation></version>\n");
	// indicate scribbles are in 72DPI resolution 
	// Precision can be improved by converting all coordinates to a higher resolution
	// since gnome_canvas stores then as floats and they are outputted as ints
	fprintf(f, "<screen><units>px</units><dpi>72</dpi></screen>\n");
	fprintf(f, "<pages>\n"); // pages
	
	page_num = 1;
    for (pagelist = journal.pages; pagelist!=NULL; pagelist = pagelist->next) 
	{
        pg = (struct Page *)pagelist->data;
		// TODO: orientation is fixed at this moment!
        fprintf(f, "<page backgroundcolor=\"#0\" id=\"%d\"><orientation>0</orientation> <height>%d</height> <width>%d</width>\n", 
				page_num, (int)pg->height, (int)pg->width);
		layer_num = 0;
        for (layerlist = pg->layers; layerlist!=NULL; layerlist = layerlist->next) 
		{
			// strokes
      		fprintf(f, "<strokes>\n");

			layer = (struct Layer *)layerlist->data;
      		for (itemlist = layer->items; itemlist!=NULL; itemlist = itemlist->next) 
			{
        		item = (struct Item *)itemlist->data;
        		if (item->type == ITEM_STROKE) 
				{
					// stroke
					fprintf(f, "<stroke color=\"#%d\"  layer=\"%d\"  penSize=\"2\" linestyle=\"solid\"  zoom=\"100.0\">\n",
						item->brush.color_no, layer_num/*, item->brush.thickness_no*/);
          			for (i=0;i<item->path->num_points;i++)
					{
            			fprintf(f, "%d %d 0\n", (int)item->path->coords[2*i], (int)item->path->coords[2*i+1]);
					}
          			fprintf(f, "</stroke>\n"); // end stoke
        		}
      		}
      		fprintf(f, "</strokes>\n"); // end stokes
			layer_num++;
    	}
    	fprintf(f, "</page>\n"); // end page
		page_num++;
  	}
	fprintf(f, "</pages>\n"); // end pages
	fprintf(f, "</notes>\n"); // end notes
		
	fclose(f);
	setlocale(LC_NUMERIC, "");

	return TRUE;
}

void
add_file_to_metadata(char *filename, char *tag)
{
	// code from notes_filestore.cpp
	gchar* dir = 0;
	gchar* file = 0;
    int ret;

	dir = g_path_get_dirname(filename);
	file = g_path_get_basename(filename);

	gchar* title = 0;
	gchar* iter = g_strrstr(file, ".xoj");
	if ( iter == 0 ) // no needle found in the haystack
	{
		title =  g_strdup(file);
	}
	else
	{
		title =  g_strndup(file, (iter - file));
	}

	// add new file to global
	//const char* mountpoint = ipc_get_mountpoint();
	erMetadb global_db = ermetadb_global_open(MOUNTPOINT_CARD, FALSE);
	if (global_db == 0) {
		//printf("failed to open globaldb.");
		return;
	}

	// date and size attributes
	struct stat statbuf;
	gchar* fname = g_strdup_printf("%s/%s", dir, file); 
	stat(fname, &statbuf);
	gint64 size = statbuf.st_size;
	gint64 mtime = statbuf.st_mtime;
	//printf("%s size %lld, mtime %lld", fname, size, mtime);
	g_free(fname);
    //                                                                        author      tag
	ret = ermetadb_global_add_file(global_db, dir, file, size, mtime, title, "Xournal", tag);
	if (ret != ER_OK) {
		printf("ermetadb_add_file failed");
	}

	ermetadb_close(global_db);
	
	g_free(dir);
	g_free(file);
	g_free(title);
}

void add_last_page_to_meta_data (const gchar *filename, int last_page)
{
    erMetadb        db;
    gchar           *file_path       = g_path_get_dirname(filename);
    gchar           *file_name       = g_path_get_basename(filename);
	int             rc;

	//printf("add_last_page_to_meta_data: %s %d\n", filename, last_page);
	
	// store metadata in recent files database on this storage device
	db = ermetadb_local_open( file_path, TRUE );
	
	if (db == NULL)
	{
		printf("cannot open database [%s]\n", file_path);
		// TODO:
		// ret = ER_FAIL;
	}
	else
	{
		// Update recent specific values in application data
		write_document_info_to_database(db, file_name, last_page);

		// close desktop database
		ermetadb_close(db);
	}

    if (file_path)   { g_free(file_path); }
    if (file_name)   { g_free(file_name); }
}

int get_last_page_from_meta_data (const gchar *filename)
{
    erMetadb        db;
    gchar           *file_path       = g_path_get_dirname(filename);
    gchar           *file_name       = g_path_get_basename(filename);
	int             rc;
	int             ret = ER_OK;
	int             last_page = -1;
	
	//printf("get_last_page_from_meta_data %s\n", filename);
	
	// store metadata in recent files database on this storage device
	db = ermetadb_local_open( file_path, TRUE );
	
	if (db == NULL)
	{
		printf("cannot open database [%s]\n", file_path);
		// TODO:
		// ret = ER_FAIL;
	}
	else
	{
		// Update recent specific values in application data
		last_page = read_document_info_from_database(db, file_name);

		// close desktop database
		ermetadb_close(db);
	}

    if (file_path)   { g_free(file_path); }
    if (file_name)   { g_free(file_name); }
	
	return last_page;
}

static void write_document_info_to_database (erMetadb db, const gchar *filename, int last_page)
{
    gboolean            ok = TRUE;
    int                 rc;
    const char          *cp;
    metadata_table      *values = NULL;
    metadata_table      *names  = NULL;
    const metadata_cell *cell   = NULL;

	//printf("entry:%s page:%d\n", filename, last_page);
    g_return_if_fail(db);

    // set new metadata for item
    //   create values table
    values = metadata_table_new();
    if (values == NULL)
    {
        ok = FALSE;
    }
    //   set recent used count
    if (ok)
    {
        rc = metadata_table_add_column(values, METADATA_KEY_XOURNAL_LAST_PAGE);
        if (rc != ER_OK)
        {
            ok = FALSE;
			printf("FAIL:could not add to names:%d\n", rc);
        }
        else
        {
            cp = g_strdup_printf( "%d", last_page );
            metadata_table_set_text( values, 0, cp );
        }
    }
    //   write to database
    if (ok)
    {
        rc = ermetadb_local_set_application_data(db, filename, values);
        if (rc != ER_OK)
        {
			printf("cannot write last page to database for file %s:%d\n", filename, rc);
            ok = FALSE;
        }
    }
	// TODO: add date/time of using this file (might be usefull for new firmware)

    metadata_table_free(values);
}

static int read_document_info_from_database (erMetadb db, const gchar *filename)
{
    gboolean            ok = TRUE;
    int                 rc;
    metadata_table      *values = NULL;
    metadata_table      *names  = NULL;
    const metadata_cell *cell   = NULL;
	int last_page = -1;

	//printf("entry:%s\n", filename);
    g_return_if_fail(db);

    // set new metadata for item
    //   create values table
    names = metadata_table_new();
    if (names == NULL)
    {
        ok = FALSE;
    }
    //   set recent used count
    if (ok)
    {
        rc = metadata_table_add_column(names, METADATA_KEY_XOURNAL_LAST_PAGE);
        if (rc != ER_OK)
        {
            ok = FALSE;
			printf("FAIL:could not add last page:%d\n", rc);
        }
    }
    // retrieve last page from database
    if (ok)
    {
        rc = ermetadb_local_get_application_data( db,
                                                  filename,
                                                  names,
                                                  &values   );
        if (rc != ER_OK)
        {
			printf("cannot retrieve last page for file %s :%d\n", filename, rc);
            ok = FALSE;
        }
    }

    // process the retrieved data, if any
    if (ok  &&  values)
    {
        cell = (metadata_cell*) (values->cell_data->data);
        if (cell->type != METADATA_TEXT)
        {
            printf("illegal cell type [%d] for metadata key [%s]\n", cell->type, METADATA_KEY_XOURNAL_LAST_PAGE);
        }
        else
        {
            //printf("last page from database [%s]\n", cell->value.v_text->str);
			last_page = atoi(cell->value.v_text->str);
		}
		
	}

    metadata_table_free(names);
    metadata_table_free(values);
	
	return last_page;
}

unsigned long get_meminfo(void)
{
	unsigned long retval=0;
#if 0
	char tmp[256]={0x0};

	FILE *shellcommand=popen("/proc/meminfo","r");
	while(fgets(tmp,sizeof(tmp),shellcommand)!=NULL)
	{
		if(memcmp(tmp,"MemFree:",8)==0)
		{
			int	wordcount=0;
			char *delimiter=" ";
			char *p=strtok(tmp,delimiter);
			while(*p)
			{
				wordcount++;
				printf("String: %d=%s\n", wordcount, p);
				if(wordcount==2) 
				{
					retval=atol(p);
				}
				p = strtok(NULL, delimiter);
			}
		}
	}
	pclose(shellcommand);
#else
	long total_pages, page_size, free_pages;
	total_pages = sysconf(_SC_PHYS_PAGES);
	page_size = sysconf(_SC_PAGE_SIZE);
	free_pages = sysconf(_SC_AVPHYS_PAGES);
	retval = (free_pages * page_size)/1024;
    printf("Free pages(%d, %d,%d):%d\n", total_pages, free_pages, page_size, retval);
#endif
	return retval;
}

long get_file_size(char *filename)
{
	struct stat statbuf;
	stat(filename, &statbuf);
	
	return statbuf.st_size;
}

